/**
 | From http://pegasus.rutgers.edu/~elflord/cpp/list_howto/ ; a
 | comment line in the web page says: "Copyright Donovan Rebbechi
 | 2001.  This tutorial may be distributed freely if this copyright
 | notice is preserved.  Changes require authors permission.  The code
 | is no-strings attached free software, it may be redistributed,
 | modified, and used.".  The author is Donovan Rebbechi, EMail
 | address elflord@pegasus.rutgers.edu; minor modifications by MLO;
 | posted to comp.lang.c++.moderated in January 2002.
 |
 | $Id: slist.hh,v 1.2 2002/05/17 09:45:51 loreti Exp $
**/

#include <iterator>

template <typename T> class List {

private:
  struct Node {
    T      _data;
    Node * _next;
    Node(const T & x, Node * y = 0) : _data(x), _next(y) {}
  };

  Node * _root;

public:
  class const_iterator;

  class iterator
    : public std::iterator<std::forward_iterator_tag, T>
  {
    friend class List;
    friend class const_iterator;

  private:
    Node * _current;

  public:
    iterator(Node * x = 0) : _current(x) {}
    iterator(const iterator & x) : _current(x._current) {}
    iterator & operator = (const iterator & x) {
      _current = x._current;
      return *this; }
    iterator & operator ++ () {
      _current = _current->_next;
      return *this; }
    iterator operator ++ (int)  {
      iterator tmp(*this);
      _current = _current->_next;
      return tmp; }
    iterator operator + (typename List<T>::iterator::difference_type n) {
      Node * tmp = _current;
      while (tmp && n--) tmp = tmp->_next;
      return iterator(tmp); }
    iterator & operator += (typename List<T>::iterator::difference_type n) {
      while (_current && n--) _current = _current->_next;
      return *this; }
    T & operator * () {
      return _current->_data; }
    T * operator -> () {
      return &(_current->_data); }
    bool operator == (const iterator & x) const {
      return _current == x._current; }
    bool operator != (const iterator & x) const {
      return _current != x._current; }
  };                            // class iterator

  class const_iterator
    : public std::iterator<std::forward_iterator_tag, const T>
  {

  private:
    const Node * _current;

  public:
    const_iterator(const Node * x = 0) : _current(x) {}
    const_iterator(const const_iterator & x) : _current(x._current) {}
    const_iterator(const iterator & x) : _current(x._current) {}
    const_iterator & operator = (const const_iterator & x) {
      _current = x._current;
      return *this; }
    const_iterator & operator = (const iterator & x) {
      _current = x._current;
      return *this; }
    const_iterator & operator ++ () {
      _current = _current->_next;
      return *this; }
    const_iterator operator ++ (int) {
      const_iterator tmp(*this);
      _current = _current->_next;
      return tmp; }
    const_iterator operator + (typename List<T>::iterator::difference_type n) {
      const Node * tmp = _current;
      while (tmp && n--) tmp = tmp->_next;
      return const_iterator(tmp); }
    const_iterator & operator += (
      typename List<T>::iterator::difference_type n) {
      while (_current && n--) _current = _current->_next;
      return *this; }
    const T & operator * () const {
      return _current->_data; }
    const T * operator -> () const {
      return &(_current->_data); }
    bool operator == (const const_iterator & x) const {
      return _current == x._current; }
    bool operator != (const const_iterator & x) const {
      return _current != x._current; }
  };                            // class const_iterator

  List() : _root(0) {}

  List(const List & rhs) : _root(0) {
    for (const_iterator i = rhs.begin();  i != rhs.end();  ++i) push_front(*i);
    reverse();
  }

  void reverse() {
    Node * last = 0;
    Node * current = _root;
    Node * next;

    while (current) {
      next = current->_next;
      current->_next = last;
      last = current;
      current = next;
    }
    _root = last;
  }

  void swap(List & x) {
    Node * tmp = _root;
    _root = x._root;
    x._root = tmp;
  }

  List & operator = (const List & x) {
    List tmp(x);
    swap(tmp);
    return *this;
  }

  ~List() { clear(); }

  void clear() { while (!empty()) pop_front(); }

  void push_front(const T & x) {
    Node * tmp = new Node(x);
    tmp->_next = _root;
    _root = tmp;
  }

  void pop_front() {
    if (_root) {
      Node * next = _root->_next;
      delete _root;
      _root = next;
    }
  }

  bool empty() { return _root; }

  T & front() { return *begin(); }

  const T & front() const { return *begin(); }

  iterator begin() { return iterator(_root); }

  iterator end() { return iterator(); }

  const_iterator begin() const { return const_iterator(_root); }

  const_iterator end() const { return const_iterator(); }

  void erase_after (iterator x) {
    Node * tmp = x._current->_next;
    if (x._current->_next) x._current->_next = x._current->_next->_next;
    delete tmp;
  }

  void insert_after (iterator x, const T & y) {
    Node * tmp = new Node(y, x._current->_next);
    x._current->_next = tmp;
  }
};                              // class List
